@@ -8,19 +8,22 @@ import shelve
 import sys
 import threading
 import traceback
-import multiprocessing
-from datetime import datetime
-from UserDict import UserDict
+try:
+    import multiprocessing
+except ImportError:
+    multiprocessing = None
 
-from celery import log
-from celery import conf
+from datetime import datetime
+
+from celery import __version__
 from celery import platforms
 from celery import registry
-from celery.execute import send_task
-from celery.schedules import maybe_schedule
-from celery.messaging import establish_connection, TaskPublisher
-from celery.utils import instantiate
-from celery.utils.info import humanize_seconds
+from celery import signals
+from celery.app import app_or_default
+from celery.log import SilenceRepeated
+from celery.schedules import maybe_schedule, crontab
+from celery.utils import cached_property, instantiate, maybe_promise
+from celery.utils.timeutils import humanize_seconds
 
 
 class SchedulingError(Exception):
@@ -72,12 +75,9 @@ class ScheduleEntry(object):
         self.args = args
         self.kwargs = kwargs
         self.options = options
-        self.last_run_at = last_run_at or self._default_now()
+        self.last_run_at = last_run_at or datetime.now()
         self.total_run_count = total_run_count or 0
 
-    def _default_now(self):
-        return datetime.now()
-
     def next(self, last_run_at=None):
         """Returns a new instance of the same class, but with
         its date and count fields updated."""
@@ -115,7 +115,7 @@ class ScheduleEntry(object):
                                                    self.schedule)
 
 
-class Scheduler(UserDict):
+class Scheduler(object):
     """Scheduler for periodic tasks.
 
     :keyword schedule: see :attr:`schedule`.
@@ -136,19 +136,31 @@ class Scheduler(UserDict):
 
     """
     Entry = ScheduleEntry
-    Publisher = TaskPublisher
 
     def __init__(self, schedule=None, logger=None, max_interval=None,
-            lazy=False, **kwargs):
-        UserDict.__init__(self)
+            app=None, Publisher=None, lazy=False, **kwargs):
         if schedule is None:
             schedule = {}
-        self.data = schedule
-        self.logger = logger or log.get_default_logger(name="celery.beat")
+        self.app = app_or_default(app)
+        conf = self.app.conf
+        self.data = maybe_promise(schedule)
+        self.logger = logger or self.app.log.get_default_logger(
+                                                name="celery.beat")
         self.max_interval = max_interval or conf.CELERYBEAT_MAX_LOOP_INTERVAL
+        self.Publisher = Publisher or self.app.amqp.TaskPublisher
         if not lazy:
             self.setup_schedule()
 
+    def install_default_entries(self, data):
+        entries = {}
+        if self.app.conf.CELERY_TASK_RESULT_EXPIRES:
+            if "celery.backend_cleanup" not in data:
+                entries["celery.backend_cleanup"] = {
+                        "task": "celery.backend_cleanup",
+                        "schedule": crontab("0", "4", "*"),
+                        "options": {"expires": 12 * 3600}}
+        self.update_from_dict(entries)
+
     def maybe_due(self, entry, publisher=None):
         is_due, next_time_to_run = entry.is_due()
 
@@ -171,19 +183,13 @@ class Scheduler(UserDict):
 
         """
         remaining_times = []
-        connection = establish_connection()
-        publisher = self.Publisher(connection=connection)
         try:
-            try:
-                for entry in self.schedule.itervalues():
-                    next_time_to_run = self.maybe_due(entry, publisher)
-                    if next_time_to_run:
-                        remaining_times.append(next_time_to_run)
-            except RuntimeError:
-                pass
-        finally:
-            publisher.close()
-            connection.close()
+            for entry in self.schedule.itervalues():
+                next_time_to_run = self.maybe_due(entry, self.publisher)
+                if next_time_to_run:
+                    remaining_times.append(next_time_to_run)
+        except RuntimeError:
+            pass
 
         return min(remaining_times + [self.max_interval])
 
@@ -217,10 +223,10 @@ class Scheduler(UserDict):
         return result
 
     def send_task(self, *args, **kwargs):               # pragma: no cover
-        return send_task(*args, **kwargs)
+        return self.app.send_task(*args, **kwargs)
 
     def setup_schedule(self):
-        pass
+        self.install_default_entries(self.data)
 
     def sync(self):
         pass
@@ -233,28 +239,45 @@ class Scheduler(UserDict):
         self.schedule[entry.name] = entry
         return entry
 
+    def _maybe_entry(self, name, entry):
+        if isinstance(entry, self.Entry):
+            return entry
+        return self.Entry(**dict(entry, name=name))
+
     def update_from_dict(self, dict_):
-        self.update(dict((name, self.Entry(name, **entry))
-                            for name, entry in dict_.items()))
+        self.schedule.update(dict((name, self._maybe_entry(name, entry))
+                                for name, entry in dict_.items()))
 
     def merge_inplace(self, b):
-        A, B = set(self.keys()), set(b.keys())
+        schedule = self.schedule
+        A, B = set(schedule.keys()), set(b.keys())
 
         # Remove items from disk not in the schedule anymore.
         for key in A ^ B:
-            self.pop(key, None)
+            schedule.pop(key, None)
 
         # Update and add new items in the schedule
         for key in B:
-            entry = self.Entry(**dict(b[key]))
-            if self.get(key):
-                self[key].update(entry)
+            entry = self.Entry(**dict(b[key], name=key))
+            if schedule.get(key):
+                schedule[key].update(entry)
             else:
-                self[key] = entry
+                schedule[key] = entry
 
     def get_schedule(self):
         return self.data
 
+    def set_schedule(self, schedule):
+        self.data = schedule
+
+    @cached_property
+    def connection(self):
+        return self.app.broker_connection()
+
+    @cached_property
+    def publisher(self):
+        return self.Publisher(connection=self.connection)
+
     @property
     def schedule(self):
         return self.get_schedule()
@@ -274,11 +297,21 @@ class PersistentScheduler(Scheduler):
         Scheduler.__init__(self, *args, **kwargs)
 
     def setup_schedule(self):
-        self._store = self.persistence.open(self.schedule_filename)
-        self.data = self._store
-        self.merge_inplace(conf.CELERYBEAT_SCHEDULE)
+        self._store = self.persistence.open(self.schedule_filename,
+                                            writeback=True)
+        if "__version__" not in self._store:
+            self._store.clear()   # remove schedule at 2.2.2 upgrade.
+        entries = self._store.setdefault("entries", {})
+        self.merge_inplace(self.app.conf.CELERYBEAT_SCHEDULE)
+        self.install_default_entries(self.schedule)
+        self._store["__version__"] = __version__
         self.sync()
-        self.data = self._store
+        self.logger.debug("Current schedule:\n" +
+                          "\n".join(repr(entry)
+                                    for entry in entries.itervalues()))
+
+    def get_schedule(self):
+        return self._store["entries"]
 
     def sync(self):
         if self._store is not None:
@@ -298,22 +331,22 @@ class Service(object):
     scheduler_cls = PersistentScheduler
 
     def __init__(self, logger=None,
-            max_interval=conf.CELERYBEAT_MAX_LOOP_INTERVAL,
-            schedule=conf.CELERYBEAT_SCHEDULE,
-            schedule_filename=conf.CELERYBEAT_SCHEDULE_FILENAME,
-            scheduler_cls=None):
-        self.max_interval = max_interval
+            max_interval=None, schedule_filename=None,
+            scheduler_cls=None, app=None):
+        self.app = app_or_default(app)
+        self.max_interval = max_interval or \
+                            self.app.conf.CELERYBEAT_MAX_LOOP_INTERVAL
         self.scheduler_cls = scheduler_cls or self.scheduler_cls
-        self.logger = logger or log.get_default_logger(name="celery.beat")
-        self.schedule = schedule
-        self.schedule_filename = schedule_filename
+        self.logger = logger or self.app.log.get_default_logger(
+                                                name="celery.beat")
+        self.schedule_filename = schedule_filename or \
+                                    self.app.conf.CELERYBEAT_SCHEDULE_FILENAME
 
-        self._scheduler = None
         self._shutdown = threading.Event()
         self._stopped = threading.Event()
         silence = self.max_interval < 60 and 10 or 1
-        self.debug = log.SilenceRepeated(self.logger.debug,
-                                         max_iterations=silence)
+        self.debug = SilenceRepeated(self.logger.debug,
+                                     max_iterations=silence)
 
     def start(self, embedded_process=False):
         self.logger.info("Celerybeat: Starting...")
@@ -321,7 +354,9 @@ class Service(object):
             "Ticking with max interval->%s" % (
                     humanize_seconds(self.scheduler.max_interval)))
 
+        signals.beat_init.send(sender=self)
         if embedded_process:
+            signals.beat_embedded_init.send(sender=self)
             platforms.set_process_title("celerybeat")
 
         try:
@@ -348,19 +383,16 @@ class Service(object):
     def get_scheduler(self, lazy=False):
         filename = self.schedule_filename
         scheduler = instantiate(self.scheduler_cls,
-                                      schedule_filename=filename,
-                                      logger=self.logger,
-                                      max_interval=self.max_interval,
-                                      lazy=lazy)
-        if not lazy:
-            scheduler.update_from_dict(self.schedule)
+                                app=self.app,
+                                schedule_filename=filename,
+                                logger=self.logger,
+                                max_interval=self.max_interval,
+                                lazy=lazy)
         return scheduler
 
-    @property
+    @cached_property
     def scheduler(self):
-        if self._scheduler is None:
-            self._scheduler = self.get_scheduler()
-        return self._scheduler
+        return self.get_scheduler()
 
 
 class _Threaded(threading.Thread):
@@ -379,31 +411,34 @@ class _Threaded(threading.Thread):
         self.service.stop(wait=True)
 
 
-class _Process(multiprocessing.Process):
-    """Embedded task scheduler using multiprocessing."""
+if multiprocessing is not None:
+    class _Process(multiprocessing.Process):
+        """Embedded task scheduler using multiprocessing."""
 
-    def __init__(self, *args, **kwargs):
-        super(_Process, self).__init__()
-        self.service = Service(*args, **kwargs)
-        self.name = "Beat"
+        def __init__(self, *args, **kwargs):
+            super(_Process, self).__init__()
+            self.service = Service(*args, **kwargs)
+            self.name = "Beat"
 
-    def run(self):
-        platforms.reset_signal("SIGTERM")
-        self.service.start(embedded_process=True)
+        def run(self):
+            platforms.reset_signal("SIGTERM")
+            self.service.start(embedded_process=True)
 
-    def stop(self):
-        self.service.stop()
-        self.terminate()
+        def stop(self):
+            self.service.stop()
+            self.terminate()
+else:
+    _Process = None
 
 
 def EmbeddedService(*args, **kwargs):
     """Return embedded clock service.
 
     :keyword thread: Run threaded instead of as a separate process.
-        Default is ``False``.
+        Default is :const:`False`.
 
     """
-    if kwargs.pop("thread", False):
+    if kwargs.pop("thread", False) or _Process is None:
         # Need short max interval to be able to stop thread
         # in reasonable time.
         kwargs.setdefault("max_interval", 1)
